#include "mps.h"

//Returns a vector of N ITensors which give the Heisenberg Hamiltonian when multiplied together
vector<ITensor> makeHeis(IndexSiteSet ISS, Real Jz = 1.0, Real Jxy = 1.0)
{
    //Set parameters
    const int d = 2;
    const int k = 5;
    const int N = ISS.N;
    
    //This will hold N ITensors - pieces of the final Hamiltonian
    vector<ITensor> H;

    //Dummy indices that will be replaced during construction
    Index S(Site,"S",d),L(Link,"L",k),R(Link,"R",k);

    //W is the translationally invariant, 4-index tensor defining the MPO
    ITensor W(S,S.primed(),L,R); W.dat = 0.0;

    //Explicit construction of W
    W(1,1,1,1) = 1.0;    W(2,2,1,1) = 1.0;       //Ident in ul corner
    W(1,1,2,1) = Jz*0.5;  W(2,2,2,1) = -Jz*0.5;  //Jz*Sz
    W(2,1,3,1) = Jxy*0.5;                        //Jxy/2*S-
    W(1,2,4,1) = Jxy*0.5;                        //Jxy/2*S+
    W(1,1,5,2) = 0.5;    W(2,2,5,2) = -0.5;      //Sz
    W(1,2,5,3) = 1.0;                            //S+
    W(2,1,5,4) = 1.0;                            //S-
    W(1,1,5,5) = 1.0;    W(2,2,5,5) = 1.0;       //Ident in lr corner

    //Boundary tensors for open boundary conditions
    ITensor Ledge(L); Ledge.dat = 0.0;
    ITensor Redge(R); Redge.dat = 0.0;

    Ledge(5) = 1.0; Redge(1) = 1.0;


    //Create a set of link indices
    vector<Index> linkSet; linkSet.push_back(Index(Link,"junk",1));
    for(int i = 1; i < N; ++i)
    {
        linkSet.push_back(Index(Link,nameint("h",i),k));
    }

    //Replace the generic W tensor with a specific one for each site
    ITensor h = Ledge*W; 
    h.mapindex(R,linkSet[1]); 
    h.mapindex(S,ISS.si(1)); h.mapindex(S.primed(),ISS.si(1).primed());
    H.push_back(h);
    for(int i = 2; i < N; ++i)
    {
        h = W;
        h.mapindex(L,linkSet[i-1]); h.mapindex(R,linkSet[i]);
        h.mapindex(S,ISS.si(i)); h.mapindex(S.primed(),ISS.si(i).primed());
        H.push_back(h);
    }
    h = W*Redge;
    h.mapindex(L,linkSet[N-1]);
    h.mapindex(S,ISS.si(N)); h.mapindex(S.primed(),ISS.si(N).primed());
    H.push_back(h);

    return H;
}


