#ifndef MPS_H
#define MPS_H
#include "tensor.h"
#include "davidson.h"
#include <iomanip>
#include <vector>
//#include "conjugate_gradient2.h"
#include "sparse.h"

const Real Sqrt2 = 1.4142135623730950488;
const Real ISqrt2 = 0.70710678118654752440;
const Real Sqrt3 = 1.7320508075688772;
const Real ISqrt3 = 0.57735026918962584;

Real ran1();

void do_denmat_Real(const ITensor& nA, ITensor& A, ITensor& B,
	Real cutoff,int maxm, bool lamright = true, bool donormalize = false);
void do_denmat_complex(const ITensor& nA, ITensor& A, ITensor& B,
	Real cutoff,int maxm, bool lamright = true, bool donormalize = false);

inline Matrix splus()  { Matrix res(2,2); res = 0.0; res(1,2) = 1.0; return res; }
inline Matrix sminus()     { Matrix res(2,2); res = 0.0; res(2,1) = 1.0; return res; }
inline Matrix sz() { Matrix res(2,2); res = 0.5; res(2,2) = -0.5;return res; }

extern Real XXZgamma;	// parameter in front of S+ S- in XXZ model
extern Real truncerror;
extern Real svdtruncerr;
extern int specialdebug1, specialdebug2;

ITensor projup(Index s1);
ITensor projdn(Index s1);
ITensor makesz(Index s1);
ITensor makesx(Index s1);
ITensor makesy(Index s1);
ITensor makeQ(int mu, int nu, Index s1);
ITensor makeP(int i, Index s1);
ITensor makezero(Index s1);
ITensor makeunit(Index s1);
void randomunit(Real& x, Real& y, Real& z);
ITensor bondH(Index s1, Index s2);
ITensor expbondH(Real tau,Index s1, Index s2,Real enbond = 0.0);
ITensor expbondiH(Real tau,Index s1, Index s2,Real enbond = 0.0);
ITensor Swap(Index s1,Index s2);

void EnlargeA(ITensor& A1, ITensor& A2, Index li, int newm);	// Enlarge Link li

void getprodcombiners(const ITensor& A, const ITensor& B, Combiner& cA, Combiner& cB);

class IndexSiteSet
    {
    vector<Index> ssi;
    vector<Index> tti;
    vector<Index> sstti;
public:
    vector<Combiner> comb;
    IndexSiteSet *pst, *pfrom;
    int N;
    IndexSiteSet() {}
    IndexSiteSet(int nsite, int ds, int end_d = 0) : ssi(nsite+1), tti(nsite+1),
        sstti(nsite+1), comb(nsite+1), pfrom(0), N(nsite) // ds = dimension d on each site
	{
        if (end_d == 0) end_d = ds;
        
        //Put spins of an optionally different length on the ends of the chain
        ssi[1] = Index(Site,nameint("s",1),end_d);
        tti[1] = Index(Site,nameint("t",1),end_d);
        sstti[1] = Index(Site,nameint("st",1),end_d*end_d);
        comb[1] = Combiner(sstti[1],ssi[1],tti[1]);

        for(int i = 2; i <= N-1; i++)
        {
            ssi[i] = Index(Site,nameint("s",i),ds); //Site is a type declared in an enum in tensor.h
            tti[i] = Index(Site,nameint("t",i),ds);
            sstti[i] = Index(Site,nameint("st",i),ds*ds);
            comb[i] = Combiner(sstti[i],ssi[i],tti[i]);
        }

        //Put spins of an optionally different length on the ends of the chain
        ssi[N] = Index(Site,nameint("s",N),end_d);
        tti[N] = Index(Site,nameint("t",N),end_d);
        sstti[N] = Index(Site,nameint("st",N),end_d*end_d);
        comb[N] = Combiner(sstti[N],ssi[N],tti[N]);

        pst = new IndexSiteSet(nsite,sstti);
        pst->pfrom = this;
	}
    IndexSiteSet(int nsite, vector<Index>& si) : ssi(nsite+1), pst(0), N(nsite) 	 // used only for MPO->MPS
	{
        for(int i = 1; i <= N; i++)
            ssi[i] = si[i];
	}

    Index si(int i) const { return ssi[i]; }
    Index ti(int i) const { return tti[i]; }
    Index sti(int i) const { return sstti[i]; }
    
    ITensor unit(int i) const
	{
	return ITensor(si(i),ti(i),1.0);
	}

    ITensor sz(int i) const
	{
        ITensor res(si(i),ti(i));
        if (si(i).m() == 2){
            //Returns the spin operator, not the pauli matrix
            res(1,1) = 0.5; res(2,2) = -0.5;
        } else if (si(i).m() == 3) {
            res(1,1) = 1.0; res(2,2) = 0.0; res(3,3) = -1.0;
        } else {
            cerr << "Spin size was " << si(i).m() << " and i was " << i << endl;
            error("Spin size not implemented in sz.");
        }
        return res;
	}

    ITensor sx(int i) const
	{
        ITensor res(si(i),ti(i));
        if (si(i).m() == 2){
            //Returns the spin operator, not the pauli matrix
            res(1,2) = 0.5; res(2,1) = 0.5;
        } else if (si(i).m() == 3) {
            res(1,2) = ISqrt2; res(2,1) = ISqrt2; res(2,3) = ISqrt2; res(3,2) = ISqrt2;
        } else {
            error("Spin size not implemented in sx.");
        }
        return res;
	}
    ITensor sy(int i) const
	{
        ITensor res(si(i),ti(i));
        if (si(i).m() == 2){
            //Returns the spin operator, not the pauli matrix
            res(1,2) = -0.5; res(2,1) = 0.5;
        } else if (si(i).m() == 3) {
            res(1,2) = -ISqrt2; res(2,1) = ISqrt2; res(2,3) = -ISqrt2; res(3,2) = ISqrt2;
        } else {
            error("Spin size not implemented in sy.");
        }
        return res * Complex_i;
	}
    ITensor splus(int i) const
	{
        ITensor res(si(i),ti(i));
        if (si(i).m() == 2){
            res(1,2) = 1.0;
        } else if (si(i).m() == 3) {
            res(1,2) = Sqrt2; res(2,3) = Sqrt2;
        } else {
            error("Spin size not implemented in splus.");
        }
        return res;
	}
    ITensor sminus(int i) const
	{
        ITensor res(si(i),ti(i));
        if (si(i).m() == 2){
            res(2,1) = 1.0;
        } else if (si(i).m() == 3) {
            res(2,1) = Sqrt2; res(3,2) = Sqrt2;
        } else {
            error("Spin size not implemented in sminus.");
        }
        return res;
	}

    ITensor Q(int mu, int nu, int i) const
	{
        ITensor res(si(i),ti(i));
        if (si(i).m() == 2) //Spin half
        {
            return res * Complex_1; //Q is zero for spin 1/2
        }
        else if (si(i).m() == 3)
        {
            if (mu == 1 && nu == 1){
                res(1,1) = -1.0/6.0;
                res(1,3) = 0.5;
                res(2,2) = 1.0/3.0;
                res(3,1) = 0.5;
                res(3,3) = -1.0/6.0;
                return res * Complex_1;
            } else if ((mu == 1 && nu == 2) || (mu == 2 && nu == 1)) {
                res(1,3) = -0.5;
                res(3,1) = 0.5;
                return res * Complex_i;
            } else if ((mu == 1 && nu == 3) || (mu == 3 && nu == 1)) {
                res(1,2) = 0.5*ISqrt2;
                res(2,1) = 0.5*ISqrt2;
                res(2,3) = -0.5*ISqrt2;
                res(3,2) = -0.5*ISqrt2;
                return res * Complex_1;
            } else if (mu == 2 && nu == 2) {
                res(1,1) = -1.0/6.0;
                res(1,3) = -0.5;
                res(2,2) = 1.0/3.0;
                res(3,1) = -0.5;
                res(3,3) = -1.0/6.0;
                return res * Complex_1;
            } else if ((mu == 2 && nu == 3) || (mu == 3 && nu == 2)) {
                res(1,2) = -0.5*ISqrt2;
                res(2,1) = 0.5*ISqrt2;
                res(2,3) = 0.5*ISqrt2;
                res(3,2) = -0.5*ISqrt2;
                return res * Complex_i; 
            } else if (mu == 3 && nu == 3) {
                res(1,1) = 1.0/3.0;
                res(2,2) = -2.0/3.0;
                res(3,3) = 1.0/3.0;
                return res * Complex_1;
            } else {
                error("Indices for Q must range between 1 and 3.");
            }
        } else {
            error("Spin size not implemented in Q.");
        }
	}

    ITensor projup(int i) const
	{
        assert(si(i).m() == 2);	// Need to implement other cases
        ITensor res(si(i),ti(i)); res(1,1) = 1.0;
        return res;
	}
    ITensor projdn(int i) const
	{
        assert(si(i).m() == 2);	// Need to implement other cases
        ITensor res(si(i),ti(i)); res(2,2) = 1.0;
        return res;
	}
    
    //Take an ITensor bham that is a combination of IndexSiteSet::sx,sy,sz etc. and 
    //makes a Trotter gate by computing exp(-tau*bham) and storing it in G
    void makeTGate(ITensor& bham, ITensor& G, Real tau)
    {
        bool is_real = false;

        Index S(Site,"S"),T(Site,"T");
        Combiner cS(S),cT(T);
        for(int i = 1; i <= N; ++i)
        {
            if(bham.findindex(si(i)) > 0) cS.addleft(S,si(i));
            if(bham.findindex(ti(i)) > 0) 
            {
                //applygate assumes that gate tensors map s indices to s' indices
                bham.mapindex(ti(i),si(i).primed());
                cT.addleft(T,si(i).primed());
            }
        }

        ITensor bTensor = cS * bham * cT;

        int n = S.m();
        if(T.m() != n) error("Bond Hamiltonian should have the same number of primed and unprimed site indices.");

        ITensor bTr,bTi;
        if(bTensor.findindex(IndReIm) > 0)
        {
            is_real = false;
            bTensor.SplitReIm(bTr,bTi);
        }
        else
        {
            is_real = true;
            bTr = bTensor;
            bTi = ITensor(S,T); bTi.dat = 0;
        }

        Vector d;
        Matrix Mre(n,n), Mim(n,n), Ure, Uim;
        bTr.getMatrix11(S,T,Mre); Mre *= exp(bTr.logfac); //getMatrix11 leaves out the logfac
        bTi.getMatrix11(S,T,Mim); Mim *= exp(bTi.logfac); //getMatrix11 leaves out the logfac
        ComplexEigenvalues(Mre,Mim,d,Ure,Uim);

        //Do the exponentiation
        for(int j = 1; j <= n; ++j)
        {
            d(j) = exp(-tau*d(j));
        }

        //Turn the matrices back into ITensors
        Index inner(Link,"inner",n);
        ITensor UUre(S,inner); UUre.fromMatrix11(S,inner,Ure);
        ITensor UUim(S,inner); UUim.fromMatrix11(S,inner,Uim);
        ITensor D(inner); D.dat = d;

        //Construct the gate with fat indices
        ITensor UU = Complex_1 * UUre + Complex_i * UUim;
        ITensor UUt = conj(UU); UUt.mapindex(S,T); //Take the Hermitian conjugate
        ITensor expbTensor = UU * (D / UUt); // '/' is noncontracting product

        if(is_real) //only keep the real part
        {
            expbTensor.SplitReIm(bTr,bTi);
            expbTensor = bTr;
        }

        //Un-combine the fat indices to form G and return
        G = cS * expbTensor * cT;
        //Acid test: This should print 'diff is 0' when the exponentiation line above is commented out
        //G.dat *= -1.0; G += bham;
        //cout << "diff is " << Norm(G.dat) << endl;
        return;
    }

}; //End class IndexSiteSet

enum direction {fromright,fromleft};
class MPS
{
public:
    int N;
    int posi;
    int maxm;
    Real cutoff;
    vector<ITensor> A;
    IndexSiteSet *Iss;
    static bool showSVDeigs;
    Index si(int i) const { return Iss->si(i); }
    Index ti(int i) const { return Iss->ti(i); }
    int NN() const { return N;}
    MPS() {}
    virtual ~MPS() {}
    MPS(IndexSiteSet& iss,int maxmm, Real cut) 
		: N(iss.N), maxm(maxmm), cutoff(cut), A(iss.N+1), Iss(&iss)
	{
        vector<Index> a(N+1);
        for(int i = 1; i <= N; i++)
            a[i] = Index(Link,nameint("a",i),1);
        A[1] = ITensor(si(1),a[1]);
        A[N] = ITensor(si(N),a[N-1]);
        for(int i = 2; i < N; i++)
            A[i] = ITensor(a[i-1],si(i),a[i]);
        for(int i = 1; i <= N; i++)
            A[i].dat.Randomize();
        posi = 1;
	}

    MPS(IndexSiteSet& iss,vector<int> initState,int maxmm, Real cut) 
		: N(iss.N), maxm(maxmm), cutoff(cut), A(iss.N+1), Iss(&iss)
	{
        vector<ITensor> a(N+1);
        for(int i = 1; i <= N; i++)
        {
            A[i] = ITensor(si(i),0.0);
            A[i](initState[i]) = 1.0;

            //Make (trivial) Link index ITensors
            a[i] = ITensor(Index(Link,nameint("a",i),1));
            a[i](1) = 1.0;
        }

        A[1] = A[1] / a[1];
        A[N] = a[N-1] / A[N];
        for(int i = 2; i < N; i++)
            A[i] = a[i-1] / A[i] / a[i];
        posi = 1;
	}

    Index LinkInd(int i)
	{
	return index_in_common(A[i],A[i+1],Link);
	}

// Standard config:  center is A[posi] A[posi+1],  right side of A[posi] is ortho
    void position(int i)
	{
	if(posi < i)
	    {
	    while(posi < i)
		{
		posi++;
		ITensor AA = A[posi] * A[posi+1];
		doSVD(posi,AA,fromleft,false);
		}
	    //cout << "moved to position " << i << endl;
	    return;
	    }
	while(posi > i)
	    {
	    ITensor AA = A[posi] * A[posi+1];
	    doSVD(posi,AA,fromright,false);
	    posi--;
	    }
	ITensor AA = A[posi] * A[posi+1];
	doSVD(posi,AA,fromleft,false);
	//cout << "moved to position " << i << endl;
	}
    void fixends()  // make first and last A[i] equal to identity matrix
	{
        ITensor unit1 = makeunit(si(1));
        ITensor unitN = makeunit(si(N));
        int m1 = si(1).m(), mN = si(N).m();
        ITensor a1 = A[1] * A[2];
        Index ia1(Link,nameint("a",1),m1);
        a1.mapindex(si(1),ia1);
        unit1.mapindex(si(1).primed(),ia1);
        A[2] = a1;  A[1] = unit1;
        ITensor aN = A[N] * A[N-1];
        Index iaN(Link,nameint("a",N-1),mN);
        aN.mapindex(si(N),iaN);
        unitN.mapindex(si(N).primed(),iaN);
        A[N-1] = aN;  A[N] = unitN;
	}
    void applygate(ITensor gate,bool donorm = false)
	{
        ITensor AA = A[posi] * A[posi+1] * gate;
        AA.noprime();
        doSVD(posi,AA,fromleft,donorm);
	}
    Real localmeas(int site,ITensor op)
	{
        int p = site-1;
        if(site == 1) p = 1;
        position(p);
        ITensor AA = A[p] * A[p+1];
        ITensor AAo = AA * op;
        AAo.noprime();
        return RDot(AAo, AA);
	}
    virtual void doSVD(int i,ITensor AA, direction dir,bool donormalize)
	{
	Real aanorm = 0;
	if(specialdebug1) aanorm = AA.norm();
	void psiphi(MPS& psi, MPS& phi,Real& re, Real& im,Real& lfac);  // <psi | phi>
	if(specialdebug1)
	    {
	    Real re,im,lfac;
	    psiphi(*this, *this,re, im,lfac);  // <psi | phi>
	    cout << "total norm before: " << re * exp(lfac) <<endl;
	    }
	if(AA.findindex(IndReIm) > 0)
	    do_denmat_complex(AA,A[i],A[i+1],cutoff,maxm,dir==fromleft,donormalize);
	else
	    do_denmat_Real(AA,A[i],A[i+1],cutoff,maxm,dir==fromleft,donormalize);
	truncerror = svdtruncerr;
	if(specialdebug1)
	    {
	    ITensor ch = A[i] * A[i+1];
	    Real chnorm = ch.norm();
	    cout << "checking doSVD norm: " << donormalize SP aanorm SP chnorm << endl;
	    Real re,im,lfac;
	    psiphi(*this, *this,re, im,lfac);  // <psi | phi>
	    cout << "total norm after: " << re * exp(lfac) <<endl;
	    }
	}

    const ITensor& AA(int i) const { return A[i]; }
    const ITensor& AAc(int i) const { return A[i]; }
    ITensor& AA(int i) 
	{ 
	return A[i]; 
	}

    void newindices()
	{
	vector<Index> a(N+1),na(N+1);
	for(int i = 1; i < N; i++)
	    a[i] = LinkInd(i),
	    na[i] = Index(Link,nameint("a",i),a[i].m());
	A[1].mapindex(a[1],na[1]);
	A[N].mapindex(a[N-1],na[N-1]);
	for(int i = 2; i < N; i++)
	    A[i].mapindex(a[i-1],na[i-1],a[i],na[i]);
	}

    MPS& operator*=(Real a)
    {
        A[posi] *= a;
        return *this;
    }
};

class MPO : public MPS
    {
public:
    bool doroot2;
    MPO() {}
    MPO(IndexSiteSet& iss, int maxmm = 5, Real cut = 1.0e-10, int initialdim = 5) 
		: MPS(iss,maxmm,cut),  doroot2(true)
	{
	maxm = maxmm; 
	cutoff = cut;
	vector<Index> hind(N+1);
	for(int i = 1; i <= N; i++)
	    hind[i] = Index(Link,nameint("Hind",i),initialdim);
	A[1] = ITensor(si(1),ti(1),hind[1]);
	A[N] = ITensor(si(N),ti(N),hind[N-1]);
	for(int i = 2; i < N; i++)
	    A[i] = ITensor(hind[i-1],si(i),ti(i),hind[i]);
	if(initialdim == 1)
	    {
	    A[1](1,1,1) = 1.0;
	    A[1](2,2,1) = 1.0;
	    A[N](1,1,1) = 1.0;
	    A[N](2,2,1) = 1.0;
	    for(int i = 2; i < N; i++)
		A[i](1,1,1,1) = 1.0, A[i](1,2,2,1) = 1.0;
	    }
	posi = 1;
	}
    void doSVD(int i,ITensor AA, direction dir,bool donormalize)
	{
	MPS::doSVD(i,AA,dir,donormalize);
	}
    };


void MPO_to_MPS(const MPO& M, MPS& m);
void MPS_to_MPO(const MPS& M, MPO& m);
    /*
ITensor bondH(Index s1, Index s2);
ITensor expbondH(Real tau,Index s1, Index s2,Real enbond = 0.0);
ITensor expbondiH(Real tau,Index s1, Index s2,Real enbond = 0.0);
ITensor Swap(Index s1,Index s2);
*/

class HamBuilder
    {
public:
    IndexSiteSet *Iss;
    int N;
    Index si(int i) { return Iss->si(i); }
    Index ti(int i) { return Iss->ti(i); }
    vector<Index> currentlinks;
    vector<ITensor> linktensor;

    HamBuilder(IndexSiteSet &iss) : Iss(&iss), N(iss.N) { }
    void newlinks()
	{
	currentlinks.resize(N);
	linktensor.resize(N);
	static int ver = 0;
	ver++;
	for(int i = 1; i < N; i++)
	    {
	    stringstream ss;
	    ss << "aham" << ver << "-" << i;
	    currentlinks[i] = Index(Link,ss.str(),1);
	    linktensor[i] = ITensor(currentlinks[i]);
	    linktensor[i](1) = 1.0;
	    }
	}

    void getidentity(MPO& res, Real factor)		// Identity
	{
	newlinks();
	res = MPO(*Iss,100,1.0e-10,1);
	res.A[1] = Iss->unit(1) * linktensor[1];
	res.A[1].dat *= factor;
	res.A[N] = Iss->unit(N) * linktensor[N-1];
	for(int i = 2; i < N; i++)
	    res.A[i] = Iss->unit(i) * linktensor[i-1] * linktensor[i];
	}
    void getMPO(MPO& res, Real factor, int i, ITensor op)
	{
	getidentity(res,1.0);
	if(i == 1)
	    res.A[i] = op * linktensor[1];
	else if(i == N)
	    res.A[i] = op * linktensor[N-1];
	else
	    res.A[i] = op * linktensor[i-1] * linktensor[i];
	res.A[1].dat *= factor;
	}
    MPO returnMPO(Real factor, int i, ITensor op)
	{
        MPO res;
        getMPO(res,factor,i,op);
        return res;
	}
    void getMPO(MPO& res, Real factor, int i, ITensor op, int i2, ITensor op2)
	{
	assert(i != i2);
	getMPO(res,1.0,i2,op2);
	if(i == 1)
	    res.A[i] = op * linktensor[1];
	else if(i == N)
	    res.A[i] = op * linktensor[N-1];
	else
	    res.A[i] = op * linktensor[i-1] * linktensor[i];
	res.A[1].dat *= factor;
	}
    };

inline ostream& operator<<(ostream& s, const MPO& m)
    {
    for(int i = 1; i <= m.N; i++)
	s << m.A[i] << endl;
    return s;
    }



void psiphi(MPS& psi, MPS& phi,Real& re, Real& im,Real& lfac);  // <psi | phi>
void psiphi(MPS& psi, MPS& phi,Real& re, Real& im);  // <psi | phi> incl logfac
void psiphi(MPO& psi, MPO& phi,Real& re, Real& im,Real& lfac);  // <psi | phi>
void psiHphi(MPS& psi, MPO& H, MPS& phi,Real& re, Real& im,Real& lfac);  //<psi|H|phi>
void psiHphi(MPS& psi, MPO& H, MPS& phi,Real& re, Real& im);  //<psi|H|phi> incl logfac
void lancstep(MPS& xi, MPO& H, MPS& xim1,Real alpha , Real beta, MPS& xip1);  
void fitset(vector<MPS>& xi, Vector& coefs, MPS& res);
void Kstep(MPS& xi, MPO& H, Real epsilon , MPS& res);  
void applyMPOfit(MPS& x, MPO& K, MPS& res);  // apply using fitting; 
void napplyMPO(MPS& x, MPO& K, MPS& res); 
void multMPO(MPO& A, MPO& Borig, MPO& res,Real cut, int m);
void nmultMPO(MPO& A, MPO& Borig, MPO& res,Real cut, int m);
void fitmposet(vector<MPO>& xi, Vector& coefs, MPO& res,Real cut, int m, bool useres = false);

void svdcut(const MatrixRef& A, Matrix& U, Vector& d, Matrix& V, Real truncerr, int maxm);
void nfitset(vector<MPS>& xxi, Vector& coefs, MPS& res,Real cut, int maxm);
void nfitset(vector<MPO>& xxi, Vector& coefs, MPO& res,Real cut, int maxm);

void expsmallH(const MPO& H, MPO& K, Real tau, Real Etot,Real Kcutoff, bool realtime);
void expH(const MPO& H, MPO& K, Real tau, Real Etot,Real Kcutoff, int ndoub,bool realtime);


class Vec2
{
public:
    Real x,y;
    Vec2(Real xx, Real yy) : x(xx), y(yy) {}
    Vec2()		// Random unit vector
	{
	do  {
	    x = (ran1()-0.5)*2.0; y = (ran1()-0.5)*2.0; 
	    } while(x*x+y*y > 1.0);
	normalize();
	}
    Vec2 normalize()
	{
	Real f = 1.0 / sqrt(x*x+y*y);
	x *= f; y *= f;
	return *this;
	}
    Real operator[](const int i) const
    {
        if (i == 1) return x;
        else if (i == 2) return y;
        else error("Index overrun in Vec2 operator().");
    }
    Real cos(){ return x; }
    Real sin(){ return y; }

    Real halfCos() //Returns |cos(theta/2)| if the angle of this Vec2 is theta.
    {
        return sqrt(0.5*(1+x));
    }
    Real halfSin() //Returns |sin(theta/2)| if the angle of this Vec2 is theta.
    {
        return sqrt(0.5*(1-x));
    }

    void from_pair(pair<Real,Real> p) { x = p.first; y = p.second;}
    pair<Real,Real> pair() const { return make_pair(x,y); }
};

inline MPS operator+(const MPS& A, const MPS& B)
{
    vector<MPS> xx(3); xx[1] = A; xx[2] = B;
    Vector co(2); co = 1.0;
    MPS res;
    nfitset(xx,co,res,A.cutoff,A.maxm);
    return res;
}
inline MPS operator-(const MPS& A, const MPS& B)
{
    MPS BB(B);
    BB *= -1.0;
    return A + BB;
}
inline MPS operator+=(MPS& A, const MPS& B)
{
    A = A + B;
    return A;
}


class Vec3
{
public:
    Real x,y,z;
    Vec3(Real xx, Real yy, Real zz) : x(xx), y(yy), z(zz) {}
    Vec3()		// Random unit vector
	{
	do  {
	    x = (ran1()-0.5)*2.0; y = (ran1()-0.5)*2.0; z = (ran1()-0.5)*2.0;
	    } while(x*x+y*y+z*z > 1.0);
	normalize();
	}
    Vec3 operator/(const Vec3& o) const	// cross product!
	{
	return Vec3(y*o.z - z*o.y,z*o.x - x * o.z, x * o.y - y * o.x);
	}
    Vec3 normalize()
	{
	Real f = 1.0 / sqrt(x*x+y*y+z*z);
	x *= f; y *= f; z *= f;
	return *this;
	}
    Real operator[](const int i) const
    {
        if (i == 1) return x;
        else if (i == 2) return y;
        else if (i == 3) return z;
        else error("Index overrun in Vec3 operator().");
    }
};

class Measure
    {
public:
    vector<Real> dat;
    Measure() {}
    void putin(Real x) { dat.push_back(x); }
    Real ave() const
	{
        int n = dat.size();
        if(n == 0) return 0.0;
        Real s = 0.0;
        for(int i = 0; i < n; i++)
            s += dat[i];
        return s/n;
	}
    Real sigma() const
	{
        int n = dat.size();
        if(n < 2) return 0.0;
        Real av = ave(), s = 0.0;
        for(int i = 0; i < n; i++)
            s += sqr(dat[i] - av);
        return sqrt(s/n);
	}
    Real err() const
	{
        Real n = dat.size();
        if(n > 1)
            return sigma()/sqrt(n-1);
        else
            return sigma()/sqrt(n);
	}
    Real binerr(int bs) const
	{
        int n = dat.size();
        if(n < bs) return 0.0;
        int nbin = n / bs;
        Vector bin(nbin);
        bin = 0.0;
        for(int i = 0; i < n; i++)
        {
            if(i/bs > (nbin-1)) break;
            bin.el(i/bs) += dat[i];
        }
        bin *= 1.0/bs;
        Measure bmeas;
        for(int i = 1; i <= nbin; i++)
            bmeas.putin(bin(i));
        return bmeas.err();
	}
    Real bootstrap(int nresamples = 1000)
    { //Uses the bootstrap procedure to estimate the std deviation
        Measure boot;
        int n = dat.size();
        if(n < 2) return 0.0;
        for(int sample = 1; sample <= nresamples; ++sample)
        {
            Real avg = 0.0;
            for(int i = 1; i <= n; ++i)
            {
                int which = int(ran1()*n);
                avg += dat[which];
            }
            avg /= n;
            boot.putin(avg);
        }
        return boot.sigma();
    }

    Real bootstrap_2c(Real prefactor,Measure firstc, bool correlated = true, int nresamples = 1000)
    { //Uses the bootstrap procedure to estimate the second cumulant error
        Measure boot2c;
        Real n = dat.size();
        if(n < 2) return 0.0;
        if(firstc.dat.size() != n) error("Measurements don't have the same size data sets.");
        for(int sample = 1; sample <= nresamples; ++sample)
        {
            Real avg2 = 0.0;
            Real avg = 0.0;
            int which;
            for(int i = 1; i <= n; ++i)
            { //Sample with replacement (replacement means 'which' can take the same value more than once)
                which = int(ran1()*n);
                avg2 += dat[which];
                if(!correlated) which = int(ran1()*n); //use <H^2> and <H> from the same METTS
                avg += firstc.dat[which];
            }
            avg2 /= n; avg /= n;
            boot2c.putin(prefactor*(avg2-avg*avg));
        }
        return boot2c.sigma();
    }


};

class ILocalHam : public BigMatrix	// to do DMRG using an MPO
{
public:
    const ITensor &LeftTerm; 
    const ITensor &RightTerm;
    const ITensor &MPOTerm;
    mutable ITensor psi;
    Vector diag;
    bool atedge;		// if true, just use LeftTerm on either edge

    ILocalHam(const ITensor& le, const ITensor& ri, const ITensor& mpo, ITensor& p) 
	: LeftTerm(le), RightTerm(ri), MPOTerm(mpo),psi(p) , atedge(false)
	{ diag.ReDimension(psi.dat.Length()); diag = 1.0; }
    ILocalHam(const ITensor& le,  const ITensor& mpo, ITensor& p) 
	: LeftTerm(le), RightTerm(le), MPOTerm(mpo),psi(p) , atedge(true)
	{ diag.ReDimension(psi.dat.Length()); diag = 1.0; }

    int Size() const { return psi.dat.Length(); }

    VectorRef DiagRef() const { return diag; }

    Vector operator *(const VectorRef &A) const
	{
        Vector res(Size());
        product(A,res);
        return res;
	}
    void product(const VectorRef &A , VectorRef & B) const
	{
        psi.AssignFromVec(A);
        ITensor psip = (atedge ? LeftTerm * psi * MPOTerm
                : LeftTerm * psi * MPOTerm * RightTerm);
        psip.mapprime(1,0);
        psi.Assign(psip);
        psi.AssignToVec(B);
	}
};

inline MPO operator+(const MPO& A, const MPO& B)
{
    vector<MPO> xx(3); xx[1] = A; xx[2] = B;
    Vector coeff(2); coeff = 1.0;
    MPO res;
    nfitset(xx,coeff,res,1.0E-14,500);
    return res;
}
inline MPO operator-(const MPO& A, const MPO& B)
{
    MPO BB(B);
    BB *= -1.0;
    return A + BB;
}
inline MPO& operator+=(MPO& A, const MPO& B)
{
    A = A+B;
    return A;
}

Real dmrg(MPS& psi, const MPO& finalham, const Sweeps& sweeps);

void measure_state(const MPS& psi,vector<Real>& sx,vector<Real>& sy,vector<Real>& sz);

ITensor makeBra(Index s, int b, Vec2 Theta, Vec2 Phi, Vec2 Omega);
ITensor makeBra(Index s, int b, Vec2 Theta, Vec2 Phi, Vec2 Omega, Vec2 Alpha);
ITensor makeiMPO(string name, int d, Index& L, Index& S, vector<Real> couplings);
MPO makeMPO(IndexSiteSet&, string name, vector<Real> couplings);

#ifdef THIS_IS_MAIN
Real XXZgamma = 1.0;
bool MPS::showSVDeigs = false;
Real truncerror = 0.0;
Real svdtruncerr = 0.0;
int specialdebug1 = 0, specialdebug2 = 0;
#endif
#endif
